package maze.searchers;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Random;

import org.junit.*;

public class MinHeapTest {
	
	public final static int NUM_ITEMS = 1000;
	
	private Comparator<Integer> c = (i, j) -> i < j ? -1 : i > j ? 1 : 0;
	
	private MinHeap<Integer> heap;
	
	@Before
	public void setup() {
		heap = new MinHeap<>(c);
	}
	
	@Test
	public void testParent() {
		assertEquals(0, MinHeap.parent(1));
		assertEquals(0, MinHeap.parent(2));
		assertEquals(1, MinHeap.parent(3));
		assertEquals(1, MinHeap.parent(4));
		assertEquals(2, MinHeap.parent(5));
		assertEquals(2, MinHeap.parent(6));
		assertEquals(3, MinHeap.parent(7));
		assertEquals(3, MinHeap.parent(8));
	}
	
	@Test
	public void testLeft() {
		assertEquals(1, MinHeap.left(0));
		assertEquals(3, MinHeap.left(1));
		assertEquals(5, MinHeap.left(2));
		assertEquals(7, MinHeap.left(3));
		assertEquals(9, MinHeap.left(4));
		assertEquals(11, MinHeap.left(5));
		assertEquals(13, MinHeap.left(6));
	}
	
	@Test
	public void testRight() {
		assertEquals(2, MinHeap.right(0));
		assertEquals(4, MinHeap.right(1));
		assertEquals(6, MinHeap.right(2));
		assertEquals(8, MinHeap.right(3));
		assertEquals(10, MinHeap.right(4));
		assertEquals(12, MinHeap.right(5));
		assertEquals(14, MinHeap.right(6));
	}
	
	@Test
	public void testLegal() {
		assertFalse(heap.legal(-1));
		assertFalse(heap.legal(0));
		for (int i = 0; i < 10; i++) {
			heap.backdoorAdd(i);
			assertTrue(heap.legal(i));
			assertFalse(heap.legal(i+1));
		}
	}
	
	@Test
	public void testIsHeap1() {
		assertTrue(heap.isHeap());
		for (int i = 0; i < 10; i++) {
			heap.add(i);
			assertTrue(heap.isHeap());
		}
	}
	
	@Test
	public void testIsHeap2() {
		assertTrue(heap.isHeap());
		heap.backdoorAdd(2);
		heap.backdoorAdd(1);
		assertFalse(heap.isHeap());
	}
	
	@Test
	public void testIsHeap3() {
		assertTrue(heap.isHeap());
		heap.backdoorAdd(1);
		heap.backdoorAdd(2);
		heap.backdoorAdd(0);
		assertFalse(heap.isHeap());
	}
	
	@Test
	public void testIsHeap4() {
		assertTrue(heap.isHeap());
		heap.backdoorAdd(0);
		heap.backdoorAdd(3);
		heap.backdoorAdd(1);
		heap.backdoorAdd(2);
		assertFalse(heap.isHeap());
	}
	
	@Test
	public void testSwap() {
		heap.backdoorAdd(10);
		heap.backdoorAdd(20);
		heap.backdoorAdd(30);
		heap.swap(0, 2);
		assertTrue(30 == heap.get(0));
		assertTrue(20 == heap.get(1));
		assertTrue(10 == heap.get(2));
		
		heap.swap(1, 2);
		assertTrue(30 == heap.get(0));
		assertTrue(10 == heap.get(1));
		assertTrue(20 == heap.get(2));
	}
	
	@Test
	public void testAdd1() {
		int[] adds = new int[]{10, 9, 11};
		int[] peeks = new int[]{10, 9, 9};
		assertEquals(adds.length, peeks.length);
		
		for (int i = 0; i < adds.length; i++) {
			heap.add(adds[i]);
			assertEquals(i + 1, heap.size());
			assertTrue(peeks[i] == heap.peek());
			assertTrue(heap.isHeap());
			allPresentInHeap(adds, i);
		}
	}
	
	public void allPresentInHeap(int[] values, int limit) {
		for (int i = 0; i <= limit; i++) {
			assertTrue(presentInHeap(values[i]));
		}
	}
	
	public boolean presentInHeap(int value) {
		for (int i = 0; i < heap.size(); i++) {
			if (value == heap.get(i)) {
				return true;
			}
		}
		return false;
	}
	
	@Test
	public void testAdd2() {
		Random rand = new Random();
		for (int i = 0; i < NUM_ITEMS; i++) {
			int oldSize = heap.size();
			int adding = rand.nextInt();
			heap.add(adding);
			assertTrue(heap.isHeap());
			assertEquals(oldSize + 1, heap.size());
		}
	}
	
	@Test
	public void testLowestInFamily1() {
		assertTrue(heap.isHeap());
		heap.backdoorAdd(0);
		heap.backdoorAdd(1);
		heap.backdoorAdd(2);
		assertEquals(0, heap.indexOfLowestInFamily(0));
	}
	
	@Test
	public void testLowestInFamily2() {
		assertTrue(heap.isHeap());
		heap.backdoorAdd(1);
		heap.backdoorAdd(0);
		heap.backdoorAdd(2);
		assertEquals(1, heap.indexOfLowestInFamily(0));
	}
	
	@Test
	public void testLowestInFamily3() {
		assertTrue(heap.isHeap());
		heap.backdoorAdd(1);
		heap.backdoorAdd(2);
		heap.backdoorAdd(0);
		assertEquals(2, heap.indexOfLowestInFamily(0));
	}
	
	@Test
	public void testRemove() {
		testAdd2();
		
		ArrayList<Integer> sortedResults = new ArrayList<>();
		while (heap.size() > 0) {
			int oldSize = heap.size();
			int removing = heap.remove();
			sortedResults.add(removing);
			assertTrue(heap.isHeap());
			assertEquals(oldSize - 1, heap.size());
		}
		
		for (int i = 1; i < sortedResults.size(); i++) {
			assertTrue(sortedResults.get(i-1).compareTo(sortedResults.get(i)) <= 0);
		}
	}
}
